@@ -3,15 +3,20 @@ package TAP::Parser;
 use strict;
 use vars qw($VERSION @ISA);
 
-use TAP::Base                 ();
-use TAP::Parser::Grammar      ();
-use TAP::Parser::Result       ();
-use TAP::Parser::Source       ();
-use TAP::Parser::Source::Perl ();
-use TAP::Parser::Iterator     ();
-use Carp                      ();
+use TAP::Base                              ();
+use TAP::Parser::Grammar                   ();
+use TAP::Parser::Result                    ();
+use TAP::Parser::ResultFactory             ();
+use TAP::Parser::Source                    ();
+use TAP::Parser::Iterator                  ();
+use TAP::Parser::IteratorFactory           ();
+use TAP::Parser::SourceHandler::Executable ();
+use TAP::Parser::SourceHandler::Perl       ();
+use TAP::Parser::SourceHandler::File       ();
+use TAP::Parser::SourceHandler::RawTAP     ();
+use TAP::Parser::SourceHandler::Handle     ();
 
-@ISA = qw(TAP::Base);
+use Carp qw( confess );
 
 =head1 NAME
 
@@ -19,11 +24,11 @@ TAP::Parser - Parse L<TAP|Test::Harness::TAP> output
 
 =head1 VERSION
 
-Version 3.03
+Version 3.20
 
 =cut
 
-$VERSION = '3.03';
+$VERSION = '3.20';
 
 my $DEFAULT_TAP_VERSION = 12;
 my $MAX_TAP_VERSION     = 13;
@@ -37,46 +42,33 @@ END {
 }
 
 BEGIN {    # making accessors
-    foreach my $method (
-        qw(
-        _stream
-        _spool
-        _grammar
-        exec
-        exit
-        is_good_plan
-        plan
-        tests_planned
-        tests_run
-        wait
-        version
-        in_todo
-        start_time
-        end_time
-        skip_all
-        )
-      )
-    {
-        no strict 'refs';
+    @ISA = qw(TAP::Base);
 
-        # another tiny performance hack
-        if ( $method =~ /^_/ ) {
-            *$method = sub {
-                my $self = shift;
-                return $self->{$method} unless @_;
-                unless ( ( ref $self ) =~ /^TAP::Parser/ ) { # trusted methods
-                    Carp::croak("$method() may not be set externally");
-                }
-                $self->{$method} = shift;
-            };
-        }
-        else {
-            *$method = sub {
-                my $self = shift;
-                return $self->{$method} unless @_;
-                $self->{$method} = shift;
-            };
-        }
+    __PACKAGE__->mk_methods(
+        qw(
+          _iterator
+          _spool
+          exec
+          exit
+          is_good_plan
+          plan
+          tests_planned
+          tests_run
+          wait
+          version
+          in_todo
+          start_time
+          end_time
+          skip_all
+          grammar_class
+          result_factory_class
+          iterator_factory_class
+          )
+    );
+
+    sub _stream {    # deprecated
+        my $self = shift;
+        $self->_iterator(@_);
     }
 }    # done making accessors
 
@@ -120,39 +112,55 @@ The arguments should be a hashref with I<one> of the following keys:
 
 =item * C<source>
 
-This is the preferred method of passing arguments to the constructor.  To
-determine how to handle the source, the following steps are taken.
+I<CHANGED in 3.18>
 
-If the source contains a newline, it's assumed to be a string of raw TAP
-output.
+This is the preferred method of passing input to the constructor.
 
-If the source is a reference, it's assumed to be something to pass to
-the L<TAP::Parser::Iterator::Stream> constructor. This is used
-internally and you should not use it.
+The C<source> is used to create a L<TAP::Parser::Source> that is passed to the
+L</iterator_factory_class> which in turn figures out how to handle the source and
+creates a <TAP::Parser::Iterator> for it.  The iterator is used by the parser to
+read in the TAP stream.
 
-Otherwise, the parser does a C<-e> check to see if the source exists.  If so,
-it attempts to execute the source and read the output as a stream.  This is by
-far the preferred method of using the parser.
+To configure the I<IteratorFactory> use the C<sources> parameter below.
 
- foreach my $file ( @test_files ) {
-     my $parser = TAP::Parser->new( { source => $file } );
-     # do stuff with the parser
- }
+Note that C<source>, C<tap> and C<exec> are I<mutually exclusive>.
 
 =item * C<tap>
 
+I<CHANGED in 3.18>
+
 The value should be the complete TAP output.
 
+The I<tap> is used to create a L<TAP::Parser::Source> that is passed to the
+L</iterator_factory_class> which in turn figures out how to handle the source and
+creates a <TAP::Parser::Iterator> for it.  The iterator is used by the parser to
+read in the TAP stream.
+
+To configure the I<IteratorFactory> use the C<sources> parameter below.
+
+Note that C<source>, C<tap> and C<exec> are I<mutually exclusive>.
+
 =item * C<exec>
 
-If passed an array reference, will attempt to create the iterator by
-passing a L<TAP::Parser::Source> object to
-L<TAP::Parser::Iterator::Source>, using the array reference strings as
-the command arguments to L<IPC::Open3::open3|IPC::Open3>:
+Must be passed an array reference.
+
+The I<exec> array ref is used to create a L<TAP::Parser::Source> that is passed
+to the L</iterator_factory_class> which in turn figures out how to handle the
+source and creates a <TAP::Parser::Iterator> for it.  The iterator is used by
+the parser to read in the TAP stream.
+
+By default the L<TAP::Parser::SourceHandler::Executable> class will create a
+L<TAP::Parser::Iterator::Process> object to handle the source.  This passes the
+array reference strings as command arguments to L<IPC::Open3::open3|IPC::Open3>:
 
  exec => [ '/usr/bin/ruby', 't/my_test.rb' ]
 
-Note that C<source> and C<exec> are mutually exclusive.
+If any C<test_args> are given they will be appended to the end of the command
+argument list.
+
+To configure the I<IteratorFactory> use the C<sources> parameter below.
+
+Note that C<source>, C<tap> and C<exec> are I<mutually exclusive>.
 
 =back
 
@@ -160,6 +168,34 @@ The following keys are optional.
 
 =over 4
 
+=item * C<sources>
+
+I<NEW to 3.18>.
+
+If set, C<sources> must be a hashref containing the names of the
+L<TAP::Parser::SourceHandler>s to load and/or configure.  The values are a
+hash of configuration that will be accessible to to the source handlers via
+L<TAP::Parser::Source/config_for>.
+
+For example:
+
+  sources => {
+    Perl => { exec => '/path/to/custom/perl' },
+    File => { extensions => [ '.tap', '.txt' ] },
+    MyCustom => { some => 'config' },
+  }
+
+This will cause C<TAP::Parser> to pass custom configuration to two of the built-
+in source handlers - L<TAP::Parser::SourceHandler::Perl>,
+L<TAP::Parser::SourceHandler::File> - and attempt to load the C<MyCustom>
+class.  See L<TAP::Parser::IteratorFactory/load_handlers> for more detail.
+
+The C<sources> parameter affects how C<source>, C<tap> and C<exec> parameters
+are handled.
+
+See L<TAP::Parser::IteratorFactory>, L<TAP::Parser::SourceHandler> and subclasses for
+more details.
+
 =item * C<callback>
 
 If present, each callback corresponding to a given result type will be called
@@ -174,7 +210,7 @@ with the result as the argument if the C<run> method is used:
  );
 
  my $aggregator = TAP::Parser::Aggregator->new;
- foreach my $file ( @test_files ) {
+ for my $file ( @test_files ) {
      my $parser = TAP::Parser->new(
          {
              source    => $file,
@@ -192,9 +228,14 @@ be used when invoking the perl executable.
 
  my $parser = TAP::Parser->new( {
      source   => $test_file,
-     switches => '-Ilib',
+     switches => [ '-Ilib' ],
  } );
 
+=item * C<test_args>
+
+Used in conjunction with the C<source> and C<exec> option to supply a reference
+to an C<@ARGV> style array of arguments to pass to the test program.
+
 =item * C<spool>
 
 If passed a filehandle will write a copy of all parsed TAP to that handle.
@@ -211,11 +252,39 @@ allow exact synchronization.
 Subtleties of this behavior may be platform-dependent and may change in
 the future.
 
+=item * C<grammar_class>
+
+This option was introduced to let you easily customize which I<grammar> class
+the parser should use.  It defaults to L<TAP::Parser::Grammar>.
+
+See also L</make_grammar>.
+
+=item * C<result_factory_class>
+
+This option was introduced to let you easily customize which I<result>
+factory class the parser should use.  It defaults to
+L<TAP::Parser::ResultFactory>.
+
+See also L</make_result>.
+
+=item * C<iterator_factory_class>
+
+I<CHANGED in 3.18>
+
+This option was introduced to let you easily customize which I<iterator>
+factory class the parser should use.  It defaults to
+L<TAP::Parser::IteratorFactory>.
+
 =back
 
 =cut
 
-# new implementation supplied by TAP::Base
+# new() implementation supplied by TAP::Base
+
+# This should make overriding behaviour of the Parser in subclasses easier:
+sub _default_grammar_class          {'TAP::Parser::Grammar'}
+sub _default_result_factory_class   {'TAP::Parser::ResultFactory'}
+sub _default_iterator_factory_class {'TAP::Parser::IteratorFactory'}
 
 ##############################################################################
 
@@ -261,6 +330,39 @@ sub run {
     }
 }
 
+##############################################################################
+
+=head3 C<make_grammar>
+
+Make a new L<TAP::Parser::Grammar> object and return it.  Passes through any
+arguments given.
+
+The C<grammar_class> can be customized, as described in L</new>.
+
+=head3 C<make_result>
+
+Make a new L<TAP::Parser::Result> object using the parser's
+L<TAP::Parser::ResultFactory>, and return it.  Passes through any arguments
+given.
+
+The C<result_factory_class> can be customized, as described in L</new>.
+
+=head3 C<make_iterator_factory>
+
+I<NEW to 3.18>.
+
+Make a new L<TAP::Parser::IteratorFactory> object and return it.  Passes through
+any arguments given.
+
+C<iterator_factory_class> can be customized, as described in L</new>.
+
+=cut
+
+# This should make overriding behaviour of the Parser in subclasses easier:
+sub make_iterator_factory { shift->iterator_factory_class->new(@_); }
+sub make_grammar          { shift->grammar_class->new(@_); }
+sub make_result           { shift->result_factory_class->make_result(@_); }
+
 {
 
     # of the following, anything beginning with an underscore is strictly
@@ -296,75 +398,90 @@ sub run {
       EOF
     );
 
+    my @class_overrides = qw(
+      grammar_class
+      result_factory_class
+      iterator_factory_class
+    );
+
     sub _initialize {
         my ( $self, $arg_for ) = @_;
 
         # everything here is basically designed to convert any TAP source to a
-        # stream.
-        $arg_for ||= {};
+        # TAP::Parser::Iterator.
 
-        $self->SUPER::_initialize( $arg_for, \@legal_callback );
+        # Shallow copy
+        my %args = %{ $arg_for || {} };
 
-        # XXX why delete() ?
-        my $stream = delete $arg_for->{stream};
-        my $tap    = delete $arg_for->{tap};
-        my $source = delete $arg_for->{source};
-        my $exec   = delete $arg_for->{exec};
-        my $merge  = delete $arg_for->{merge};
-        my $spool  = delete $arg_for->{spool};
+        $self->SUPER::_initialize( \%args, \@legal_callback );
 
-        if ( 1 < grep {defined} $stream, $tap, $source ) {
-            $self->_croak(
-                "You may only choose one of 'stream', 'tap', or 'source'");
+        # get any class overrides out first:
+        for my $key (@class_overrides) {
+            my $default_method = "_default_$key";
+            my $val = delete $args{$key} || $self->$default_method();
+            $self->$key($val);
         }
-        if ( $source && $exec ) {
+
+        my $iterator = delete $args{iterator};
+        $iterator ||= delete $args{stream};    # deprecated
+        my $tap         = delete $args{tap};
+        my $raw_source  = delete $args{source};
+        my $sources     = delete $args{sources};
+        my $exec        = delete $args{exec};
+        my $merge       = delete $args{merge};
+        my $spool       = delete $args{spool};
+        my $switches    = delete $args{switches};
+        my $ignore_exit = delete $args{ignore_exit};
+        my $test_args   = delete $args{test_args} || [];
+
+        if ( 1 < grep {defined} $iterator, $tap, $raw_source, $exec ) {
             $self->_croak(
-                '"source" and "exec" are mutually exclusive options');
+                "You may only choose one of 'exec', 'tap', 'source' or 'iterator'"
+            );
         }
+
+        if ( my @excess = sort keys %args ) {
+            $self->_croak("Unknown options: @excess");
+        }
+
+        # convert $tap & $exec to $raw_source equiv.
+        my $type   = '';
+        my $source = TAP::Parser::Source->new;
         if ($tap) {
-            $stream = TAP::Parser::Iterator->new( [ split "\n" => $tap ] );
+            $type = 'raw TAP';
+            $source->raw( \$tap );
         }
         elsif ($exec) {
-            my $source = TAP::Parser::Source->new;
-            $source->source($exec);
-            $source->merge($merge);    # XXX should just be arguments?
-            $stream = $source->get_stream;
+            $type = 'exec ' . $exec->[0];
+            $source->raw( { exec => [ @$exec, @$test_args ] } );
         }
-        elsif ($source) {
-            if ( my $ref = ref $source ) {
-                $stream = TAP::Parser::Iterator->new($source);
-            }
-            elsif ( -e $source ) {
-
-                my $perl = TAP::Parser::Source::Perl->new;
-
-                $perl->switches( $arg_for->{switches} )
-                  if $arg_for->{switches};
-
-                $perl->merge($merge);    # XXX args to new()?
-
-                $stream = $perl->source_file($source)->get_stream;
-            }
-            else {
-                $self->_croak("Cannot determine source for $source");
-            }
+        elsif ($raw_source) {
+            $type = 'source ' . ref($raw_source) || $raw_source;
+            $source->raw( ref($raw_source) ? $raw_source : \$raw_source );
+        }
+        elsif ($iterator) {
+            $type = 'iterator ' . ref($iterator);
         }
 
-        unless ($stream) {
-            $self->_croak( 'PANIC: could not determine stream' );
+        if ( $source->raw ) {
+            my $src_factory = $self->make_iterator_factory($sources);
+            $source->merge($merge)->switches($switches)
+              ->test_args($test_args);
+            $iterator = $src_factory->make_iterator($source);
+        }
+
+        unless ($iterator) {
+            $self->_croak(
+                "PANIC: could not determine iterator for input $type");
         }
 
         while ( my ( $k, $v ) = each %initialize ) {
             $self->{$k} = 'ARRAY' eq ref $v ? [] : $v;
         }
 
-        $self->_stream($stream);
-        my $grammar = TAP::Parser::Grammar->new($stream);
-        $grammar->set_version( $self->version );
-        $self->_grammar($grammar);
+        $self->_iterator($iterator);
         $self->_spool($spool);
-
-        $self->start_time( $self->get_time );
+        $self->ignore_exit($ignore_exit);
 
         return $self;
     }
@@ -396,6 +513,10 @@ examples of each, are as follows:
 
  1..42
 
+=item * Pragma
+
+ pragma +strict
+
 =item * Test
 
  ok 3 - We should start with some foobar!
@@ -491,13 +612,6 @@ C<$result> object.
 
 This is merely a synonym for C<as_string>.
 
-=head3 C<tests_planned>
-
-  my $planned = $result->tests_planned;
-
-Returns the number of tests planned.  For example, a plan of C<1..17> will
-cause this method to return '17'.
-
 =head3 C<directive>
 
  my $directive = $result->directive;
@@ -513,6 +627,18 @@ If a SKIP directive is included with the plan, this method will return it.
 If a SKIP directive was included with the plan, this method will return the
 explanation, if any.
 
+=head2 C<pragma> methods
+
+ if ( $result->is_pragma ) { ... }
+
+If the above evaluates as true, the following methods will be available on the
+C<$result> object.
+
+=head3 C<pragmas>
+
+Returns a list of pragmas each of which is a + or - followed by the
+pragma name.
+ 
 =head2 C<commment> methods
 
  if ( $result->is_comment ) { ... }
@@ -774,6 +900,47 @@ This method lets you know which (or how many) tests had SKIP directives.
 
 sub skipped { @{ shift->{skipped} } }
 
+=head2 Pragmas
+
+=head3 C<pragma>
+
+Get or set a pragma. To get the state of a pragma:
+
+  if ( $p->pragma('strict') ) {
+      # be strict
+  }
+
+To set the state of a pragma:
+
+  $p->pragma('strict', 1); # enable strict mode
+
+=cut
+
+sub pragma {
+    my ( $self, $pragma ) = splice @_, 0, 2;
+
+    return $self->{pragma}->{$pragma} unless @_;
+
+    if ( my $state = shift ) {
+        $self->{pragma}->{$pragma} = 1;
+    }
+    else {
+        delete $self->{pragma}->{$pragma};
+    }
+
+    return;
+}
+
+=head3 C<pragmas>
+
+Get a list of all the currently enabled pragmas:
+
+  my @pragmas_enabled = $p->pragmas;
+
+=cut
+
+sub pragmas { sort keys %{ shift->{pragma} || {} } }
+
 =head2 Summary Results
 
 These results are "meta" information about the total results of an individual
@@ -850,10 +1017,10 @@ failed, any TODO tests unexpectedly succeeded, or any parse errors occurred.
 
 sub has_problems {
     my $self = shift;
-    return $self->failed
+    return
+         $self->failed
       || $self->parse_errors
-      || $self->wait
-      || $self->exit;
+      || ( !$self->ignore_exit && ( $self->wait || $self->exit ) );
 }
 
 =head3 C<version>
@@ -879,6 +1046,20 @@ Once the parser is done, this will return the wait status.  If the parser ran
 an executable, it returns the wait status of the executable.  Otherwise, this
 mererely returns the C<exit> status.
 
+=head2 C<ignore_exit>
+
+  $parser->ignore_exit(1);
+
+Tell the parser to ignore the exit status from the test when determining
+whether the test passed. Normally tests with non-zero exit status are
+considered to have failed even if all individual tests passed. In cases
+where it is not possible to control the exit value of the test script
+use this option to ignore it.
+
+=cut
+
+sub ignore_exit { shift->pragma( 'ignore_exit', @_ ) }
+
 =head3 C<parse_errors>
 
  my @errors = $parser->parse_errors; # the parser errors
@@ -956,14 +1137,33 @@ sub _make_state_table {
     my %state_globals = (
         comment => {},
         bailout => {},
+        yaml    => {},
         version => {
             act => sub {
-                my ($version) = @_;
                 $self->_add_error(
                     'If TAP version is present it must be the first line of output'
                 );
             },
         },
+        unknown => {
+            act => sub {
+                my $unk = shift;
+                if ( $self->pragma('strict') ) {
+                    $self->_add_error(
+                        'Unknown TAP token: "' . $unk->raw . '"' );
+                }
+            },
+        },
+        pragma => {
+            act => sub {
+                my ($pragma) = @_;
+                for my $pr ( $pragma->pragmas ) {
+                    if ( $pr =~ /^ ([-+])(\w+) $/x ) {
+                        $self->pragma( $2, $1 eq '+' );
+                    }
+                }
+            },
+        },
     );
 
     # Provides default elements for transitions
@@ -1002,12 +1202,11 @@ sub _make_state_table {
                     }
                 }
 
-                if ($number) {
+                if ( defined $number ) {
                     if ( $number != $tests_run ) {
                         my $count = $tests_run;
-                        $self->_add_error(
-                            "Tests out of sequence.  Found ($number) but expected ($count)"
-                        );
+                        $self->_add_error( "Tests out of sequence.  Found "
+                              . "($number) but expected ($count)" );
                     }
                 }
                 else {
@@ -1031,16 +1230,17 @@ sub _make_state_table {
                   } => $number;
             },
         },
-        yaml => {
-            act => sub { },
-        },
+        yaml => { act => sub { }, },
     );
 
-# Each state contains a hash the keys of which match a token type. For each token
-# type there may be:
-#   act      A coderef to run
-#   goto     The new state to move to. Stay in this state if missing
-#   continue Goto the new state and run the new state for the current token
+    # Each state contains a hash the keys of which match a token type. For
+    # each token
+    # type there may be:
+    #   act      A coderef to run
+    #   goto     The new state to move to. Stay in this state if
+    #            missing
+    #   continue Goto the new state and run the new state for the
+    #            current token
     %states = (
         INIT => {
             version => {
@@ -1056,9 +1256,9 @@ sub _make_state_table {
                     }
                     if ( $ver_num > $MAX_TAP_VERSION ) {
                         $self->_add_error(
-                            "TAP specified version $ver_num but we don't know "
-                              . "about versions later than $MAX_TAP_VERSION"
-                        );
+                                "TAP specified version $ver_num but "
+                              . "we don't know about versions later "
+                              . "than $MAX_TAP_VERSION" );
                         $ver_num = $MAX_TAP_VERSION;
                     }
                     $self->version($ver_num);
@@ -1094,8 +1294,8 @@ sub _make_state_table {
                     my ($plan) = @_;
                     my $line = $self->plan;
                     $self->_add_error(
-                        "Plan ($line) must be at the beginning or end of the TAP output"
-                    );
+                            "Plan ($line) must be at the beginning "
+                          . "or end of the TAP output" );
                     $self->is_good_plan(0);
                 },
                 continue => 'PLANNED'
@@ -1114,7 +1314,7 @@ sub _make_state_table {
     );
 
     # Apply globals and defaults to state table
-    for my $name ( sort keys %states ) {
+    for my $name ( keys %states ) {
 
         # Merge with globals
         my $st = { %state_globals, %{ $states{$name} } };
@@ -1142,21 +1342,34 @@ determine the readiness of this parser.
 
 =cut
 
-sub get_select_handles { shift->_stream->get_select_handles }
+sub get_select_handles { shift->_iterator->get_select_handles }
+
+sub _grammar {
+    my $self = shift;
+    return $self->{_grammar} = shift if @_;
+
+    return $self->{_grammar} ||= $self->make_grammar(
+        {   iterator => $self->_iterator,
+            parser   => $self,
+            version  => $self->version
+        }
+    );
+}
 
 sub _iter {
     my $self        = shift;
-    my $stream      = $self->_stream;
-    my $spool       = $self->_spool;
+    my $iterator    = $self->_iterator;
     my $grammar     = $self->_grammar;
+    my $spool       = $self->_spool;
     my $state       = 'INIT';
     my $state_table = $self->_make_state_table;
 
+    $self->start_time( $self->get_time );
+
     # Make next_state closure
     my $next_state = sub {
         my $token = shift;
         my $type  = $token->type;
-        my $count = 1;
         TRANS: {
             my $state_spec = $state_table->{$state}
               or die "Illegal state: $state";
@@ -1173,10 +1386,24 @@ sub _iter {
                     $state = $goto;
                 }
             }
+            else {
+                confess("Unhandled token type: $type\n");
+            }
         }
         return $token;
     };
 
+    # Handle end of stream - which means either pop a block or finish
+    my $end_handler = sub {
+        $self->exit( $iterator->exit );
+        $self->wait( $iterator->wait );
+        $self->_finish;
+        return;
+    };
+
+    # Finally make the closure that we return. For performance reasons
+    # there are two versions of the returned function: one that handles
+    # callbacks and one that does not.
     if ( $self->_has_callbacks ) {
         return sub {
             my $result = eval { $grammar->tokenize };
@@ -1198,11 +1425,9 @@ sub _iter {
                 print {$spool} $result->raw, "\n" if $spool;
             }
             else {
-                $self->exit( $stream->exit );
-                $self->wait( $stream->wait );
-                $self->_finish;
-
-                $self->_make_callback( 'EOF', $result );
+                $result = $end_handler->();
+                $self->_make_callback( 'EOF', $self )
+                  unless defined $result;
             }
 
             return $result;
@@ -1220,9 +1445,7 @@ sub _iter {
                 print {$spool} $result->raw, "\n" if $spool;
             }
             else {
-                $self->exit( $stream->exit );
-                $self->wait( $stream->wait );
-                $self->_finish;
+                $result = $end_handler->();
             }
 
             return $result;
@@ -1235,6 +1458,18 @@ sub _finish {
 
     $self->end_time( $self->get_time );
 
+    # Avoid leaks
+    $self->_iterator(undef);
+    $self->_grammar(undef);
+
+    # If we just delete the iter we won't get a fault if it's recreated.
+    # Instead we set it to a sub that returns an infinite
+    # stream of undef. This segfaults on 5.5.4, presumably because
+    # we're still executing the closure that gets replaced and it hasn't
+    # been protected with a refcount.
+    $self->{_iter} = sub {return}
+      if $] >= 5.006;
+
     # sanity checks
     if ( !$self->plan ) {
         $self->_add_error('No plan found in TAP output');
@@ -1256,9 +1491,8 @@ sub _finish {
         my $actual = $self->tests_run;
         my $passed = $self->passed;
         my $failed = $self->failed;
-        $self->_croak(
-            "Panic: planned test count ($actual) did not equal sum of passed ($passed) and failed ($failed) tests!"
-        );
+        $self->_croak( "Panic: planned test count ($actual) did not equal "
+              . "sum of passed ($passed) and failed ($failed) tests!" );
     }
 
     $self->is_good_plan(0) unless defined $self->is_good_plan;
@@ -1299,7 +1533,7 @@ result as its argument.
  );
 
  my $aggregator = TAP::Parser::Aggregator->new;
- foreach my $file ( @test_files ) {
+ for my $file ( @test_files ) {
      my $parser = TAP::Parser->new(
          {
              source    => $file,
@@ -1349,9 +1583,9 @@ Invoked if C<< $result->is_unknown >> returns true.
 
 =item * C<ELSE>
 
-If a result does not have a callback defined for it, this callback will be
-invoked.  Thus, if all of the previous result types are specified as callbacks,
-this callback will I<never> be invoked.
+If a result does not have a callback defined for it, this callback will
+be invoked. Thus, if all of the previous result types are specified as
+callbacks, this callback will I<never> be invoked.
 
 =item * C<ALL>
 
@@ -1392,8 +1626,8 @@ test output:
 
 =item * C<EOF>
 
-Invoked when there are no more lines to be parsed.  Since there is
-no accompanying L<TAP::Parser::Result> object the C<TAP::Parser> object is
+Invoked when there are no more lines to be parsed. Since there is no
+accompanying L<TAP::Parser::Result> object the C<TAP::Parser> object is
 passed instead.
 
 =back
@@ -1413,18 +1647,19 @@ L<Test::Harness>.  However, there are some minor differences.
 
 =item * TODO plans
 
-A little-known feature of L<Test::Harness> is that it supported TODO lists in
-the plan:
+A little-known feature of L<Test::Harness> is that it supported TODO
+lists in the plan:
 
  1..2 todo 2
  ok 1 - We have liftoff
  not ok 2 - Anti-gravity device activated
 
-Under L<Test::Harness>, test number 2 would I<pass> because it was listed as a
-TODO test on the plan line.  However, we are not aware of anyone actually
-using this feature and hard-coding test numbers is discouraged because it's
-very easy to add a test and break the test number sequence.  This makes test
-suites very fragile.  Instead, the following should be used:
+Under L<Test::Harness>, test number 2 would I<pass> because it was
+listed as a TODO test on the plan line. However, we are not aware of
+anyone actually using this feature and hard-coding test numbers is
+discouraged because it's very easy to add a test and break the test
+number sequence. This makes test suites very fragile. Instead, the
+following should be used:
 
  1..2
  ok 1 - We have liftoff
@@ -1432,7 +1667,8 @@ suites very fragile.  Instead, the following should be used:
 
 =item * 'Missing' tests
 
-It rarely happens, but sometimes a harness might encounter 'missing tests:
+It rarely happens, but sometimes a harness might encounter
+'missing tests:
 
  ok 1
  ok 2
@@ -1440,16 +1676,136 @@ It rarely happens, but sometimes a harness might encounter 'missing tests:
  ok 16
  ok 17
 
-L<Test::Harness> would report tests 3-14 as having failed.  For the
-C<TAP::Parser>, these tests are not considered failed because they've never
-run.  They're reported as parse failures (tests out of sequence).
+L<Test::Harness> would report tests 3-14 as having failed. For the
+C<TAP::Parser>, these tests are not considered failed because they've
+never run. They're reported as parse failures (tests out of sequence).
 
 =back
 
+=head1 SUBCLASSING
+
+If you find you need to provide custom functionality (as you would have using
+L<Test::Harness::Straps>), you're in luck: C<TAP::Parser> and friends are
+designed to be easily plugged-into and/or subclassed.
+
+Before you start, it's important to know a few things:
+
+=over 2
+
+=item 1
+
+All C<TAP::*> objects inherit from L<TAP::Object>.
+
+=item 2
+
+Many C<TAP::*> classes have a I<SUBCLASSING> section to guide you.
+
+=item 3
+
+Note that C<TAP::Parser> is designed to be the central 'maker' - ie: it is
+responsible for creating most new objects in the C<TAP::Parser::*> namespace.
+
+This makes it possible for you to have a single point of configuring what
+subclasses should be used, which in turn means that in many cases you'll find
+you only need to sub-class one of the parser's components.
+
+The exception to this rule are I<SourceHandlers> & I<Iterators>, but those are
+both created with customizeable I<IteratorFactory>.
+
+=item 4
+
+By subclassing, you may end up overriding undocumented methods.  That's not
+a bad thing per se, but be forewarned that undocumented methods may change
+without warning from one release to the next - we cannot guarantee backwards
+compatability.  If any I<documented> method needs changing, it will be
+deprecated first, and changed in a later release.
+
+=back
+
+=head2 Parser Components
+
+=head3 Sources
+
+A TAP parser consumes input from a single I<raw source> of TAP, which could come
+from anywhere (a file, an executable, a database, an io handle, a uri, etc..).
+The source gets bundled up in a L<TAP::Parser::Source> object which gathers some
+meta data about it.  The parser then uses a L<TAP::Parser::IteratorFactory> to
+determine which L<TAP::Parser::SourceHandler> to use to turn the raw source
+into a stream of TAP by way of L</Iterators>.
+
+If you simply want C<TAP::Parser> to handle a new source of TAP you probably
+don't need to subclass C<TAP::Parser> itself.  Rather, you'll need to create a
+new L<TAP::Parser::SourceHandler> class, and just plug it into the parser using
+the I<sources> param to L</new>.  Before you start writing one, read through
+L<TAP::Parser::IteratorFactory> to get a feel for how the system works first.
+
+If you find you really need to use your own iterator factory you can still do
+so without sub-classing C<TAP::Parser> by setting L</iterator_factory_class>.
+
+If you just need to customize the objects on creation, subclass L<TAP::Parser>
+and override L</make_iterator_factory>.
+
+Note that L</make_source> & L</make_perl_source> have been I<DEPRECATED> and
+are now removed.
+
+=head3 Iterators
+
+A TAP parser uses I<iterators> to loop through the I<stream> of TAP read in
+from the I<source> it was given.  There are a few types of Iterators available
+by default, all sub-classes of L<TAP::Parser::Iterator>.  Choosing which
+iterator to use is the responsibility of the I<siterator factory>, though it
+simply delegates to the I<Source Handler> it uses.
+
+If you're writing your own L<TAP::Parser::SourceHandler>, you may need to
+create your own iterators too.  If so you'll need to subclass
+L<TAP::Parser::Iterator>.
+
+Note that L</make_iterator> has been I<DEPRECATED> and is now removed.
+
+=head3 Results
+
+A TAP parser creates L<TAP::Parser::Result>s as it iterates through the
+input I<stream>.  There are quite a few result types available; choosing
+which class to use is the responsibility of the I<result factory>.
+
+To create your own result types you have two options:
+
+=over 2
+
+=item option 1
+
+Subclass L<TAP::Parser::Result> and register your new result type/class with
+the default L<TAP::Parser::ResultFactory>.
+
+=item option 2
+
+Subclass L<TAP::Parser::ResultFactory> itself and implement your own
+L<TAP::Parser::Result> creation logic.  Then you'll need to customize the
+class used by your parser by setting the C<result_factory_class> parameter.
+See L</new> for more details.
+
+=back
+
+If you need to customize the objects on creation, subclass L<TAP::Parser> and
+override L</make_result>.
+
+=head3 Grammar
+
+L<TAP::Parser::Grammar> is the heart of the parser - it tokenizes the TAP
+input I<stream> and produces results.  If you need to customize its behaviour
+you should probably familiarize yourself with the source first.  Enough
+lecturing.
+
+Subclass L<TAP::Parser::Grammar> and customize your parser by setting the
+C<grammar_class> parameter.  See L</new> for more details.
+
+If you need to customize the objects on creation, subclass L<TAP::Parser> and
+override L</make_grammar>
+
 =head1 ACKNOWLEDGEMENTS
 
-All of the following have helped.  Bug reports, patches, (im)moral support, or
-just words of encouragement have all been forthcoming.
+All of the following have helped. Bug reports, patches, (im)moral
+support, or just words of encouragement have all been forthcoming.
 
 =over 4
 
@@ -1487,6 +1843,12 @@ just words of encouragement have all been forthcoming.
 
 =item * Matt Kraai
 
+=item * David Wheeler
+
+=item * Alex Vandiver
+
+=item * Cosimo Streppone
+
 =back
 
 =head1 AUTHORS
@@ -1501,22 +1863,28 @@ Michael Peters <mpeters at plusthree dot com>
 
 Leif Eriksen <leif dot eriksen at bigpond dot com>
 
+Steve Purkis <spurkis@cpan.org>
+
+Nicholas Clark <nick@ccl4.org>
+
+Lee Johnson <notfadeaway at btinternet dot com>
+
 =head1 BUGS
 
 Please report any bugs or feature requests to
-C<bug-tapx-parser@rt.cpan.org>, or through the web interface at
-L<http://rt.cpan.org/NoAuth/ReportBug.html?Queue=TAP-Parser>.
-We will be notified, and then you'll automatically be notified of progress on
-your bug as we make changes.
+C<bug-test-harness@rt.cpan.org>, or through the web interface at
+L<http://rt.cpan.org/NoAuth/ReportBug.html?Queue=Test-Harness>.
+We will be notified, and then you'll automatically be notified of
+progress on your bug as we make changes.
 
-Obviously, bugs which include patches are best.  If you prefer, you can patch
-against bleed by via anonymous checkout of the latest version:
+Obviously, bugs which include patches are best. If you prefer, you can
+patch against bleed by via anonymous checkout of the latest version:
 
- svn checkout http://svn.hexten.net/tapx
+ git clone git://github.com/AndyA/Test-Harness.git
 
 =head1 COPYRIGHT & LICENSE
 
-Copyright 2006-2007 Curtis "Ovid" Poe, all rights reserved.
+Copyright 2006-2008 Curtis "Ovid" Poe, all rights reserved.
 
 This program is free software; you can redistribute it and/or modify it
 under the same terms as Perl itself.
